from math import pi, cos, sin, sqrt, atan2, log10
import cmath
from random import gauss
import sys

import numpy as np
from scipy.signal import welch, firwin, lfilter, sosfilt, butter, chirp
import matplotlib.pyplot as plt

import optparse

from wavfile_utils import input_samples
from windows import get_window


parser = optparse.OptionParser ()
parser.add_option ('--lpf', action = 'store', type = float, default = 0)
parser.add_option ('--hpf', action = 'store', type = float, default = 0)
parser.add_option ('--left', action = 'store_true', default = False)
parser.add_option ('--log', action = 'store_true', default = False)
options, args = parser.parse_args ()
assert len(args) == 1

samples, channels, SPS, bits, comment, fullsc = input_samples (args[0])


M = samples.shape
if len(M) == 2:
    if options.left:
        samples = samples [:, 0]
    else:
        samples = samples [:, M[1]-1]
    M = samples.shape

SKIP_SAMPLES = SPS
MAXSAMPLES = SPS

if M[0] > SKIP_SAMPLES + MAXSAMPLES:
    samples = samples [SKIP_SAMPLES:]
    M = samples.shape

    
M = M[0]
if M > MAXSAMPLES:
    samples = samples[0 : MAXSAMPLES]
    M = MAXSAMPLES
samples *= 1e6   # units are ppm

print ("SPS %s, bits %s, samplecount %s" % (SPS, bits, M))

if 0 < options.lpf < SPS/2:
    print ("low pass filtering to %s Hz" % options.lpf)
    brickwall = firwin (127, options.lpf, nyq = SPS/2)
    samples = lfilter (brickwall, [1], samples) [127: ]
    M -= 127


average =  np.sum(samples) / M
samples -= average

if 0 < options.hpf < 1000:
    print ("high pass filtering to %s Hz" % options.hpf)
    brickwall = firwin (1023, options.hpf, nyq = SPS/2, pass_zero = False)
    samples = lfilter (brickwall, [1], samples) [1023: ]
    M -= 1023
    

NUMBER_PASSES = 3

def rms (samples):
    #sum_sqrs2 = 0
    #for i in range (len (samples)):
    #    sum_sqrs2 += samples[i]*samples[i]
    sum_sqrs = np.sum (samples * samples)
    
    total = samples.shape[0]
    #print ("in RMS: sum %s %s, total %s, ratio %s, len %s" % (sum_sqrs, sum_sqrs2, total, sum_sqrs/total, len(samples)))
    return sqrt (sum_sqrs / total)


def count_zero_crossings (samples):
    m = samples.shape[0]
    first = None
    last = None
    count = 0
    sense = -1 if samples[0] < 0 else +1
    for i in range (m):
        if samples[i] != 0:
            if sense * samples[i] < 0:  # sign changed
                sense *= -1
                if first is None:
                    first = i
                else:
                    if count % 2 == 0:
                        last = i
                    count += 1
    print ("zero crossings in %s samples: %s %s %s" % (m, count-1, first, last))
    return count-1, first, last


BPH_MASK =  0xFFFFFFFFFFFF
BPH_TURN = 0x1000000000000

def correlate (phase, freq, array, offset = 0):
    bphase = int (round (BPH_TURN * phase / 2 / pi)) & BPH_MASK
    m = array.shape[0]
    csum = 0.0
    ssum = 0.0
    bphase_inc = int (round (BPH_TURN * freq / SPS)) & BPH_MASK
    bphase += (offset * bphase_inc) & BPH_MASK
    bphase &= BPH_MASK
    for i in range (m):
        angle = bphase * 2 * pi / BPH_TURN
        csum += cos (angle) * array [i]
        ssum += sin (angle) * array [i]
        bphase += bphase_inc
        bphase &= BPH_MASK
    return (2 * csum / m), (2 * ssum / m)




def estimate_fundamental (samples, number_passes = NUMBER_PASSES, lowpass_crossings = 1500):
    m = samples.shape[0]
    if lowpass_crossings:
        low_pass = butter (4, lowpass_crossings*2 / SPS, btype = 'lowpass', analog = False, output = 'sos')
        lsamples = sosfilt (low_pass, samples)
        n, first, last = count_zero_crossings (lsamples)
    else:
        n, first, last = count_zero_crossings (samples)
    freq_est = (0.5 * n / (last - first)) * SPS
    print ("freq now %s" % freq_est)

    halfway = m // 2
    if lowpass_crossings:
        _, _, half_last = count_zero_crossings (lsamples[0 : halfway])
        _, half_first, _ = count_zero_crossings (samples[halfway : m])
    else:
        _, _, half_last = count_zero_crossings (lsamples[0 : halfway])
        _, half_first, _ = count_zero_crossings (samples[halfway : m])
    half_first += halfway

    half_first = half_last
    print ("Got %s %s-%s %s-%s" % (n, first, half_first, half_last, last))

    #print ("Using segments %s--%s, %s--%s" % (first, half_last, half_first, last))

    phase_est = 0.0
    for passes in range (number_passes):
        x1, y1 = correlate (phase_est, freq_est, samples[first : half_last], offset = first)
        x2, y2 = correlate (phase_est, freq_est, samples[half_first : last], offset = half_first)
        x3, y3 = correlate (phase_est, freq_est, samples, offset = 0)
        x4, y4 = correlate (phase_est, freq_est, samples[first : last], offset = first)

        amp_est = sqrt (0.5 * (x1*x1 + y1*y1 + x2*x2 + y2*y2))
        amp_est2 = sqrt (x3*x3 + y3*y3)
        #print ("CORREL     %s, %s  and %s, %s (amp est %s | %s, rms %s)" % (x1, y1, x2, y2, amp_est, amp_est2, rms(samples)))
        
        ph1 = atan2 (y1, x1)
        ph2 = atan2 (y2, x2)
        if ph2 - ph1 > pi:
            print ("XXXX")
            ph2 -= 2*pi
        elif ph1 - ph2 > pi:
            print ("YYYY")
            ph1 -= 2*pi
        #print ("phases %s %s and %s, %s overall" % (ph1, ph2, atan2(y3, x3), atan2(y4, x4)))
        freq_est -= (2 * (ph2 - ph1)) * SPS / (2 * pi * m)
        phase_est -= 0.5 * (3 * ph1 - ph2)
        print ("freq now %s" % freq_est)
        print ("amp, ph now %s, %s" % (amp_est, phase_est))
        print ()
    #print ("phases %s %s" % (ph1, ph2))

    XX = amp_est * cos (phase_est)
    YY = amp_est * sin (phase_est)
    amplitude = XX + 1j*YY
    model = gen_model (amplitude, freq_est)

    # correct amplitude and phase by using the difference from model (the correlations with raw
    # signal are inaccurate because quantized in time and we don't interpolate)
    error = samples - model
    x, y = correlate (0, freq_est, error, offset = 0)  # correlate error to model, absolute phase
    amplitude += x - 1j*y            # adjust model
    amp_est = abs (amplitude)
    phase_est = cmath.phase (amplitude)
    model = gen_model (amplitude, freq_est)
    error = samples - model

    x, y = correlate (phase_est, freq_est, error, offset = 0)
    print ("diff %s, %s   %s, %s" % (x, y, atan2 (y, x), sqrt (x*x + y*y)))

    average =  np.sum(error) / M
    error -= average

    return model, error, freq_est


def OLDgen_model (amplitude, freq_est):
    model = np.zeros ([ M ], dtype=float)
    amp_est = abs (amplitude)
    phase_est = cmath.phase (amplitude)
    for i in range (M):
        model[i] = amp_est * cos (phase_est + i * 2 * pi * freq_est / SPS)
    return model


def gen_model (amplitude, freq_est):
    return abs(amplitude) * chirp (np.linspace (0, M/SPS, num = M, endpoint=False),
                                   freq_est, M/SPS, freq_est,
                                   method = 'linear',
                                   phi = cmath.phase(amplitude)*180/pi)

def gen_model_quad (amplitude, freq_est):
    return abs(amplitude) * chirp (np.linspace (0, M/SPS, num = M, endpoint=False),
                                   freq_est, M/SPS, freq_est,
                                   method = 'linear',
                                   phi = cmath.phase(amplitude)*180/pi - 90)

def wrap_phase (ph):
    while ph > pi:
        ph -= 2 * pi
    while ph < -pi:
        ph += 2 * pi
    return ph

def dB (s):
    return 10.0 * np.log10 (np.sum (s*s) / s.shape[0])


def find_component (array, freq):
    x, y = correlate (0, freq, array, offset = 0)
    return sqrt (0.5 * (x*x + y*y))
    

model, diff, freq = estimate_fundamental (samples)
#model, diff, freq = estimate_fundamental (samples, lowpass_crossings = 100)
#print ("NOTE THAT THE BANDWIDTH IS LIMITED TO 100Hz")
print ("Freq %8.4fHz" % freq)

def generate_hann_window(n):
    wind = np.zeros ([n], float)
    for i in range (n):
        wind[i] = cos (2*pi*i/n)
    return 0.5 * (1.0 - wind)

windowed_samples = generate_hann_window (M) * samples


HARM_COUNT = 31 if SPS > 65000 else 21

components = np.zeros ([HARM_COUNT], float)
power_sum = 0.0
for i in range (1, HARM_COUNT):
    if i * freq <= HARM_COUNT * 1000 + 500:
        components[i] = find_component (windowed_samples, i * freq)
        relative_amp = components[i] / components[1]
        relative_power = relative_amp * relative_amp
        if i > 1:
            power_sum += relative_power
            #print ("H%s component %5.2fdB" % (i, 10 * log10 (relative_power)))


THD = 10 * log10 (power_sum)
THD_perc = sqrt(power_sum) * 100
print ("THD = %s dB, %7.5f%%" % (THD, sqrt(power_sum) * 100))
if THD_perc > 20.0:
    print ("Not a pure tone?")
    sys.exit(4)

signal_db = dB(model)
diff_db = dB(diff)
SiNaD = signal_db - diff_db
distortion_ppm = 1000000 * rms(diff) / rms(model)
print ("SINAD %s dB, %7.5f%%" % (SiNaD, distortion_ppm / 10000.0))

SHOW_SAMPLES = 2400 #int (0.40 * SPS)

fig = plt.figure()

ax = fig.add_subplot (311)

ax.plot (np.arange(SHOW_SAMPLES)*1e3/SPS, model[M//2 : M//2 + SHOW_SAMPLES]/1e6, 'k')
#ax.plot (np.arange(SHOW_SAMPLES)*1e3/SPS, samples[M//2 : M//2 + SHOW_SAMPLES]/1e6, 'c')
#ax.plot (np.arange(SHOW_SAMPLES)*1e3/SPS, samples[0 : SHOW_SAMPLES]/1e6, 'r')
#ax.plot (np.arange(SHOW_SAMPLES)*1e3/SPS, samples[-SHOW_SAMPLES : ]/1e6, 'b')
ax.set_title ("model wave")
ax.grid()

ax = fig.add_subplot (312)

ax.plot (np.arange(SHOW_SAMPLES)*1e3/SPS, diff[M//2 : M//2 + SHOW_SAMPLES], 'k')
ax.set_title ("residual (ppm, ms)")
ax.grid()
#ax.plot (range(M//EXPAND), model[0:M//EXPAND]/3000, 'r')
#ax.plot (range(M//EXPAND), samples[0:M//EXPAND]/3000, 'b')


NPERSEG = 0x2000
WINDOW = get_window ('HFT144D', NPERSEG)

f, ns = welch (samples, nperseg = NPERSEG, return_onesided = True, window = WINDOW)
maxns = np.max (ns)
ax2 = fig.add_subplot (313)
if options.log:
    ax2.semilogx (f*SPS, 10.0 * np.log10 (ns / maxns), 'g')
    ax2.set_xlim (20, SPS/2)
else:
    ax2.plot (f*SPS, 10.0 * np.log10 (ns / maxns), 'g')

ax2.set_title ("Residual error")
ax2.set_ylim (-140, 10)
ax2.set_yticks (range (-140, 1, 10))
ax2.grid (True)
ax2.set_title ("SINAD %6.2f dB, THD %6.2f dB, %7.5f%%" % (dB(model) - dB(diff), THD, THD_perc))
ax2.set_xlabel ('fs')
ax2.set_ylabel ('dB')

for i in range (1, HARM_COUNT):
    amp = components[i] / components[1]
    if amp != 0.0:
        dB = 20 * log10 (amp)
        ax2.plot ([i * freq, i * freq], [-140, dB], 'r')

ax2.plot ([0, 0.5*SPS], [-SiNaD,-SiNaD], 'b')
ax2.plot ([0, 0.5*SPS], [THD, THD], 'r')
plt.show()
