from math import sqrt
from time import time

import numpy as np
import scipy.io.wavfile as wavfile


# change to false and comment out the imports below if PySoundFile isn't installed
USE_SOUNDFILE = True

from soundfile import SoundFile
import soundfile

LEFT = 0
RIGHT = 1


def rdint4 (f):
    data = f.read (4)
    return int (data[0]) | int (data[1]) << 8 | int (data[2]) << 16 | int (data[3]) << 24

def rdint3 (f):
    data = f.read (3)
    return int (data[0]) | int (data[1]) << 8 | int (data[2]) << 16

def rdint2 (f):
    data = f.read (2)
    return int (data[0]) | int (data[1]) << 8


def rdsample (f, bits_per_sample):
    if bits_per_sample == 24:
        val = rdint3 (f)
        if val & 0x800000 != 0:
            return (val - 0x1000000) * sqrt(2) / 8388608.0
        else:
            return val * sqrt(2) / 8388608.0
    else:
        val = rdint2 (f)
        if val & 0x8000 != 0:
            return (val - 0x10000) * sqrt(2) / 32768.0
        else:
            return val * sqrt(2) / 32768.0

def rdintsample (f, bits_per_sample):
    if bits_per_sample == 24:
        val = rdint3 (f)
        if val & 0x800000 != 0:
            return (val - 0x1000000)
        else:
            return val
    else:
        val = rdint2 (f)
        if val & 0x8000 != 0:
            return (val - 0x10000)
        else:
            return val

def rd4 (f):
    return f.read(4)



def wrint4 (f, n):
    for i in range (4):
        f.write (bytes ([n & 0xFF]))
        n >>= 8

def wrint3 (f, n):
    f.write (bytes ([n & 0xFF]))
    f.write (bytes ([(n >> 8) & 0xFF]))
    f.write (bytes ([(n >> 16) & 0xFF]))

def wrint2 (f, n):
    f.write (bytes ([n & 0xFF]))
    f.write (bytes ([n >> 8]))

def parse_raw_str (bs):
    ind = len(bs) - 1
    while ind >= 0 and bs[ind] == 0:
        ind -= 1
    return bs[0 : ind+1].decode ('UTF-8')



def find_soundfile_comment(raw):
    start = raw.find ('fullscale ')
    if start == -1:
        return None
    end = raw.find ('\n', start)
    return raw[start:end] if end != -1 else raw[start:]


def input_samples (filename, scaling = 1.0, absolute = False, need_comments = False, fullscale = 1.0):
    split_at = filename.find (':')
    if split_at > 0:
        scaling = pow (10, float (filename [split_at+1:]) / 20)
        filename = filename[0:split_at]
    before = time()
    try:
        if USE_SOUNDFILE and absolute:
            try:
                sndfile = SoundFile (filename)
                rate = sndfile.samplerate
                channels = sndfile.channels
                data = sndfile.read (always_2d = True)
                print ("Fast SoundFile read, shape %s, rate %s, format %s, subtype %s" % (data.shape, rate, sndfile.format, sndfile.subtype))
                bits = 16 if sndfile.subtype == 'PCM_16' else (24 if sndfile.subtype == 'PCM_24' else None)
                print ("bits %s" % bits)
                return sqrt(2) * scaling * data, channels, rate, bits, find_soundfile_comment (sndfile.extra_info), fullscale
            except Exception as e:
                print ("Fast SoundFile read failed %s, using slow method" % e)
              
        else:
            if absolute and not need_comments:
                try:
                    rate, data = wavfile.read (filename)
                    channels = data.shape[1]
                    print ("Fast scipy read, shape %s, rate %s" % (data.shape, rate))
                    return sqrt(2) / 32767.0 * scaling * data, channels, rate, 16, None, fullscale
                except ValueError as e:
                    print ("Fast scipy read failed %s, using slow method" % e)
        return input_samples_slow (filename, scaling = scaling, absolute = absolute, split_at = split_at, fullscale = fullscale)
    finally:
        after = time()
        print ("reading took %s seconds" % (after-before))

def input_samples_slow (filename, scaling = 1.0, absolute = False, split_at = -1, int_samples = False, fullscale = 1.0):
    print ("Slow read wavfile")
    with open (filename, 'rb') as f:
        sig = rd4 (f)
        assert sig == b'RIFF'

        lenleft = rdint4 (f) # , 4 + 8 + 16 + 8 + data_len)
        sig = rd4 (f)
        assert sig == b'WAVE'

        sig = rd4 (f)
        assert sig == b'fmt '

        chunklen = rdint4 (f)
        lenleft -= 12
        lenleft -= chunklen

        print ("chunklen = %x" % chunklen)
        assert chunklen == 16
        samptype = rdint2 (f)
        assert samptype == 1   # linear samples
        channels = rdint2 (f)
        assert channels == 1 or channels == 2   # mono or stereo
        sample_rate = rdint4 (f)
        byte_rate = rdint4 (f)
        bytes_per_sample = rdint2 (f)

        bits_per_sample = rdint2 (f)
        assert bits_per_sample == 16 or bits_per_sample == 24

        sig = rd4 (f)
        assert sig == b'data'
        data_len = rdint4 (f)
        lenleft -= 8
        print ("len left = %s ,data_len = %s" % (lenleft, data_len))
        sample_count = data_len // bytes_per_sample
        print ("channels %s, sample type %s, sample_rate %s, byte_rate %s bytes/samp %s, duration %ss, N %s" % (channels, samptype, sample_rate, byte_rate, bytes_per_sample, sample_count/sample_rate, sample_count))

        if int_samples:
            samples = np.zeros ([sample_count, channels], int)
            for i in range (sample_count):
                for c in range (channels):
                    samples [i,c] = rdintsample (f, bits_per_sample)
        else:
            samples = np.zeros ([sample_count, channels], float)
            print ("samples %s, channels %s" % (sample_count, channels))
            for i in range (sample_count):
                for c in range (channels):
                    samples [i,c] = rdsample (f, bits_per_sample)


        lenleft -= data_len
        comment = None
        while lenleft > 0:
            sig = rd4(f)
            clen = rdint4(f)
            print ("seen post data segment %s, len %s, left %s" % (sig, clen, lenleft))
            lenleft -= 8
            lenleft -= clen
            if sig == b'LIST':
                print ("list chunk")
                listtype = rd4(f)
                clen -= 4
                if listtype == b'INFO':
                    while clen > 0:
                        infotype = rd4(f)
                        subclen = rdint4(f)
                        subdata = f.read(subclen)
                        if infotype == b'ICMT':
                            comment = parse_raw_str (subdata)
                            print ("Comment from %s is '%s'" % (subdata, comment))
                            if comment.startswith ("fullscale "):
                                print ("Full scale is recorded as %s" % comment [10:])
                                fullscale = parse_comment (comment)
                        else:
                            print ("skipping info sub chunk %s, len %s" % (infotype, subclen))
                        clen -= 8 + subclen
                else:
                    f.read (clen)
            else:
                print ("skipping segment %s of len %s" % (sig, clen))
                f.read (clen)

    if int_samples:
        return samples, channels, sample_rate, bits_per_sample, comment, fullscale
        
    if split_at <= 0 and not absolute:
        rms = sqrt (np.sum (samples * samples) / len (samples))
        scaling = 1.0 / rms

    return scaling * samples, channels, sample_rate, bits_per_sample, comment, fullscale



def parse_level (s):
    s = s.strip()
    def parse_qty (eoff):
        return float (s[0:eoff])
    
    if s.endswith ("dBV"):
        return pow (10.0, parse_qty (-3) / 20.0)
    elif s.endswith ("dBu"):
        return pow (10.0, parse_qty (-3) / 20.0) * sqrt(0.6)
    elif s.endswith ("uV"):
        return 1e-6 * parse_qty (-2)
    elif s.endswith ("mV"):
        return 1e-3 * parse_qty (-2)
    elif s.endswith ("V"):
        return parse_qty (-1)
    else:
        raise ValueError ("Fullscale comment not parsable: '%s'" % s)

    
def parse_comment (comment):
    if comment.startswith ("fullscale "):
        return parse_level (comment [10:])
    else:
        return 1.0


def output_samples (filename, samples, sample_rate = 48000, bits_per_sample = 16, comment = None):
    siz = samples.shape
    channels = 1 if len(siz) == 1 else siz[1]
    print ("Channels to output %s" % channels)
    sample_count = len (samples)
    bytes_per_sample = (bits_per_sample + 7) // 8
    data_len = bytes_per_sample * channels * sample_count
    comment = comment.encode('UTF-8') if comment is not None else None
    clen = 1 + len(comment) if comment is not None else 0

    example_sample = samples[0] if len (samples.shape) == 1 else samples[0, 0]
    if type (example_sample) == np.float64:
        if bits_per_sample == 16:
            samples = np.array (32768.0 * samples, dtype = int)
        else:
            samples = np.array (256 * 32768.0 * samples, dtype = int)
    
    with open (filename, 'wb') as f:
        f.write (b'RIFF')
        wrint4 (f, 4 + 8 + 16 + 8 + data_len + (clen + 20 if clen != 0 else 0))
        f.write (b'WAVE')

        f.write (b'fmt ')
        wrint4 (f, 16)
        wrint2 (f, 1)  # linear samples
        wrint2 (f, channels)
        wrint4 (f, sample_rate)  # samples/second
        wrint4 (f, bytes_per_sample * channels * sample_rate)  # bytes/second
        wrint2 (f, bytes_per_sample * channels)
        wrint2 (f, bits_per_sample)

        f.write (b'data')
        wrint4 (f, data_len)

        if len(samples.shape) == 1:
            for i in range (sample_count):
                if bits_per_sample == 24:
                    wrint3 (f, 0xFFFFFF & samples [i])
                elif bits_per_sample == 16:
                    wrint2 (f, 0xFFFF & samples [i])
        else:
            print ("outputing %s samples for %s channels" % (sample_count, channels))
            for i in range (sample_count):
                for c in range (channels):
                    if bits_per_sample == 24:
                        wrint3 (f, 0xFFFFFF & samples [i, c])
                    elif bits_per_sample == 16:
                        wrint2 (f, 0xFFFF & samples [i, c])

        if clen != 0:
            f.write (b'LIST')
            wrint4 (f, clen + 12)
            f.write (b'INFO')
            f.write (b'ICMT')
            wrint4 (f, clen)
            f.write (comment)
            f.write (b'\0')


def tag_file (from_file, to_file, comment):
    samples, channels, rate, bits, orig_comment, orig_fullscale = input_samples_slow (from_file, int_samples = True)
    output_samples (to_file, samples, sample_rate = rate, bits_per_sample = bits, comment = comment)


def swap_LR (from_file, to_file):
    samples, channels, rate, bits, orig_comment, orig_fullscale = input_samples_slow (from_file, int_samples = True)
    samples = samples [:, ::-1]
    output_samples (to_file, samples, sample_rate = rate, bits_per_sample = bits)

def mono_from_L (from_file, to_file):
    samples, channels, rate, bits, orig_comment, orig_fullscale = input_samples_slow (from_file, int_samples = True)
    sh = samples.shape
    assert sh[1] == 2
    new_samples = np.zeros ([sh[0], 2], int)
    new_samples [:, LEFT] = samples [:, LEFT]
    new_samples [:, RIGHT] = samples [:, LEFT]
    output_samples (to_file, new_samples, sample_rate = rate, bits_per_sample = bits)

def mono_from_R (from_file, to_file):
    samples, channels, rate, bits, orig_comment, orig_fullscale = input_samples_slow (from_file, int_samples = True)
    sh = samples.shape
    assert sh[1] == 2
    new_samples = np.zeros ([sh[0], 2], int)
    new_samples [:, LEFT] = samples [:, RIGHT]
    new_samples [:, RIGHT] = samples [:, RIGHT]
    output_samples (to_file, new_samples, sample_rate = rate, bits_per_sample = bits)
